/* Agent Tcl
   Bob Gray
   10 January 1996

   tclLocation.cc

   This file reimplements class LOCATION for use in the Tcl/Tk interpreters.

   Copyright (c) 1995-1996, Bob Gray, Dartmouth College

   See the file "agent.terms" for information on usage and redistribution
   of this file and for a DISCLAIMER OF ALL WARRANTIES.
*/

#ifndef NO_PRAGMAS
#pragma implementation
#endif

#include "platPorting.h"
#include "agentMachId.h"      // MachineId class
#include "genDString.h"

/* ------------------------------------------------------------------ */

#include "tcl.h"
#include "genConversion.h"
#include "tclLocation.h"
#include "truefalse.h"

static char *agentTrace (ClientData clientData, Tcl_Interp *interp, char *name1, char *name2, int flags);

/* LOCATION_TCL::LOCATION_TCL

   Purpose: This procedure is the constructor for class LOCATION_TCL.

     Input: p_agent = agent associated with this LOCATION_TCL instance
		      (class AGENT_TCL *)
*/

LOCATION_TCL::LOCATION_TCL (AGENT_TCL *p_agent): LOCATION (p_agent)
{
    interp       = NULL;
    masterInterp = NULL;
    modified     = FALSE;
}

/* LOCATION_TCL::setInterp

   Purpose: Set the Tcl interpreter

     Input: interp       = the Tcl interpreter associated with this instance
		           (struct Tcl_Interp *)

	    masterInterp = the first interpreter's master
			   (struct Tcl_Interp *)
*/

void LOCATION_TCL::setInterp (Tcl_Interp *newInterp, Tcl_Interp *newMasterInterp)
{
    int flags = TCL_GLOBAL_ONLY | TCL_TRACE_WRITES | TCL_TRACE_UNSETS;

	/* assertions on the parameters */

    assert (interp == NULL);
    assert (masterInterp == NULL);
    assert (newInterp != NULL);

	/* remember the interpreter */

    interp       = newInterp;
    masterInterp = newMasterInterp;  

	/* load the Tcl array and turn on the traces */
 
    refresh ();

    Tcl_TraceVar2 (interp, "agent", NULL, flags, agentTrace, (ClientData) this);  
    Tcl_GlueVar2 (interp, "agent", NULL, TCL_GLOBAL_ONLY);

    if (masterInterp != NULL) {
	Tcl_TraceVar2 (masterInterp, "agent", NULL, flags, agentTrace, (ClientData) this);
	Tcl_GlueVar2 (masterInterp, "agent", NULL, TCL_GLOBAL_ONLY);
    }
}

/* LOCATION_TCL::~LOCATION_TCL

   Purpose: This procedure is the destructor for class LOCATION_TCL.
*/

LOCATION_TCL::~LOCATION_TCL ()
{
    int flags = TCL_GLOBAL_ONLY | TCL_TRACE_WRITES | TCL_TRACE_UNSETS;

	/* turn off the variable trace */

    Tcl_UntraceVar2 (interp, "agent", (char *) NULL, flags, agentTrace, (ClientData) this);
    Tcl_UnsetVar2 (interp, "agent", (char *) NULL, TCL_GLOBAL_ONLY);

    if (masterInterp != NULL) {
	Tcl_UntraceVar2 (masterInterp, "agent", (char *) NULL, flags, agentTrace, (ClientData) this);
	Tcl_UnsetVar2 (masterInterp, "agent", (char *) NULL, TCL_GLOBAL_ONLY);
    }
}

/* LOCATION_TCL::refresh_string

   Purpose: Reload an element of the "agent" array

     Input: name   = name of the array element
		     (char *)

            value  = new value for the array element
		     (char *)

    Output: The procedure reloads the element. 
*/

void LOCATION_TCL::refresh_string (char *name, char *value)
{
	/* allow modification */

    modified = TRUE;

	/* do the agent interpreter */

    Tcl_SetVar2 (interp, "agent", name, (value == NULL) ? "" : value, TCL_GLOBAL_ONLY);

	/* do the master interpreter */

    if (masterInterp != NULL) {
	Tcl_SetVar2 (masterInterp, "agent", name, (value == NULL) ? "" : value, TCL_GLOBAL_ONLY);
    }

	/* disallow modification */

    modified = FALSE;
}

/* LOCATION_TCL::refresh_version

   Purpose: Refresh the version section of the "agent" array

     Input: None

    Output: The procedure refreshes the version section of the "agent" array.
*/

void LOCATION_TCL::refresh_version (void)
{
    char temp[32];
    Version version;

	/* reload */

    version = agent -> getCoreVersion ();
    sprintf (temp, "%d %d %s", version.m_major, version.m_minor, (version.m_type == Version::e_ALPHA) ? "alpha" : (version.m_type == Version::e_BETA) ?  "beta" : "production");
    refresh_string ("core-version", temp); 

    version = agent -> getInterpVersion ();
    sprintf (temp, "%d %d %s", version.m_major, version.m_minor, (version.m_type == Version::e_ALPHA) ? "alpha" : (version.m_type == Version::e_BETA) ?  "beta" : "production");
    refresh_string ("version", temp); 
    refresh_string ("language", version.m_name.value());
}

/* LOCATION_TCL::refresh

   Purpose: Reload the "agent" array

     Input: None

    Output: The procedure reloads the agent array.
*/

void LOCATION_TCL::refresh (void)
{
    char *string;
    char temp[32];

	/* reload */

    if (!(isRegistered())) {

	refresh_string ("root-server"    , (char *) NULL);
	refresh_string ("root-ip"        , (char *) NULL);
	refresh_string ("root-name"      , (char *) NULL);
	refresh_string ("root-id"        , (char *) NULL);
	refresh_string ("root"           , (char *) NULL);
        refresh_string ("roothome-server", (char *) NULL);
        refresh_string ("roothome-ip"    , (char *) NULL);
        refresh_string ("roothome-name"  , (char *) NULL);
        refresh_string ("roothome-id"    , (char *) NULL);
        refresh_string ("roothome"       , (char *) NULL);
	refresh_string ("local-server"   , (char *) NULL);
	refresh_string ("local-ip"       , (char *) NULL);
	refresh_string ("local-name"     , (char *) NULL);
	refresh_string ("local-id"       , (char *) NULL);
	refresh_string ("local"          , (char *) NULL);
	refresh_string ("home-server"    , (char *) NULL);
	refresh_string ("home-ip"        , (char *) NULL);
	refresh_string ("home-name"      , (char *) NULL);
	refresh_string ("home-id"        , (char *) NULL);
	refresh_string ("home"           , (char *) NULL);

    } else {

	refresh_string ("root-server", rootId.getServerName().value());
        refresh_string ("roothome-server", rootHomeId.getServerName().value());
	refresh_string ("home-server", homeId.getServerName().value());
	refresh_string ("local-server", localId.getServerName().value());
	refresh_string ("root-name", rootId.getName().value());
        refresh_string ("roothome-name", rootHomeId.getName().value());
	refresh_string ("home-name", homeId.getName().value());
	refresh_string ("local-name", localId.getName().value());
    
	sprintf (temp, "%u", rootId.getId()); 
	refresh_string ("root-id", temp);

        sprintf (temp, "%u", rootHomeId.getId());
 	refresh_string ("roothome-id", temp);

	sprintf (temp, "%u", homeId.getId());
	refresh_string ("home-id", temp);

	sprintf (temp, "%u", localId.getId()); 
	refresh_string ("local-id", temp);

        string = Agent_IpToString (rootId.getServerIp());
	refresh_string ("root-ip", string);
	delete string;

        string = Agent_IpToString (rootHomeId.getServerIp());
	refresh_string ("roothome-ip", string);
	delete string;

	string = Agent_IpToString (homeId.getServerIp());
	refresh_string ("home-ip", string);
	delete string;

        string = Agent_IpToString (localId.getServerIp());
	refresh_string ("local-ip", string);
	delete string;

	string = Agent_IdToString (&rootId);
	refresh_string ("root", string);
	delete string;

	string = Agent_IdToString (&rootHomeId);
	refresh_string ("roothome", string);
	delete string;

	string = Agent_IdToString (&homeId);
	refresh_string ("home", string);
	delete string;

	string = Agent_IdToString (&localId);
	refresh_string ("local", string);
	delete string;
    }

	/* reload the temp directory, actual machine and the registration flags */

    refresh_string ("temp-directory", tempDirectory.value());
    refresh_string ("actual-server", machine.getName().value());

    string = Agent_IpToString (machine.getIp());
    refresh_string ("actual-ip", string);
    delete string;

    sprintf (temp, "%d", (registered == AGENT::e_REGISTERED) ? 1 : 0);
    refresh_string ("registered", temp);

    sprintf (temp, "%d", (root == AGENT::e_ROOT_AGENT) ? 1 : 0);
    refresh_string ("isRoot", temp);

    sprintf (temp, "%d", (home == AGENT::e_AT_HOME) ? 1 : 0);
    refresh_string ("isHome", temp);

    sprintf (temp, "%d", (server == AGENT::e_INCOMING_AGENT) ? 1 : 0);
    refresh_string ("server", temp);

    sprintf (temp, "%d", TCL_MAX_PREC);
    refresh_string ("maxprecision", temp);

	/* refresh the version information */

    refresh_version ();
}

/* LOCATION_TCL::changeNotification

   Purpose: Callback for when the agent's location information has changed
*/

void LOCATION_TCL::changeNotification (void)
{
    refresh ();
}

/* agentTrace

   Purpose: This procedure traces the "agent" array and prevents writes and
	    unsets.
*/

static char *agentTrace (ClientData clientData, Tcl_Interp *interp, char *, char *, int flags)
{
    int tflags = TCL_GLOBAL_ONLY | TCL_TRACE_WRITES | TCL_TRACE_UNSETS;

	/* LOCATION instance */

    LOCATION_TCL *location = (LOCATION_TCL *) clientData;
    assert (location != NULL);

	/* allow the operation if modified flag is SET */

    if (location -> allowModification()) {
      return ((char *) NULL);
    }

	/* restore original values on a write */

    if (flags & TCL_TRACE_WRITES) {
	Tcl_UnsetVar2 (interp, "agent", NULL, TCL_GLOBAL_ONLY);
	location -> refresh ();
	return "can not overwrite agent identification";
    }

	/* otherwise we are trying to unset                 */

    if (!(flags & TCL_INTERP_DESTROYED)) {

	location -> refresh ();

	if (flags & TCL_TRACE_DESTROYED) {
	    Tcl_TraceVar2 (interp, "agent", NULL, tflags, agentTrace, (ClientData) location);  
	}

	Tcl_GlueVar2 (interp, "agent", NULL, TCL_GLOBAL_ONLY);
	return "can not unset agent identification";
    }

    return ((char *) NULL);
}
