/* Agent Tcl
   Bob Gray
   26 June 1995

   genConnection.cc

   This file implements the classes that handle message-oriented connections.

   Copyright (c) 1995, Robert S. Gray Dartmouth College

   See the file "agent.terms" for information on usage and redistribution
   of this file and for a DISCLAIMER OF ALL WARRANTIES.
*/

#ifndef NO_PRAGMAS
#pragma implementation
#endif

#include "platPorting.h"
#include "platExclusion.h"
#include "agentMachId.h"
#include "genConnection.h"
#include "genDBuffer.h"
#include "genDString.h"
#include "genEncrypt.h"
#include "genMessage.h"
#include "platSigio.h"
#include "genTcpip.h"
#include "genTransmit.h"

Connection::Connection (const BASE_SECURITY &security):
    m_security (security),
    m_status (Connection::e_INITIALIZED)
{
    // empty
}

Connection::~Connection () {
    // empty
}

AsyncConnection::AsyncConnection (const BASE_SECURITY &security):
    Connection (security)
{
    // empty
}

AsyncConnection::~AsyncConnection (void) {
    // empty
}

SocketConnection::SocketConnection (const BASE_SECURITY &security):
    AsyncConnection (security),
    m_trap (NULL),
    m_machine (),
    m_handler (NULL),
    m_sockfd (-1)
{
    // empty
}

SocketConnection::SocketConnection (int sockfd, const MachineId &machine, const BASE_SECURITY &security):
    AsyncConnection (security),
    m_trap (NULL),
    m_machine (machine),
    m_handler (NULL),
    m_sockfd (sockfd)
{
    FDMASK_ZERO (m_set);
    FDMASK_SET (m_sockfd, m_set);
    m_index = FDMASK_INDEX (m_sockfd);
    Connection::setStatus (Connection::e_CONNECTED);
}

SocketConnection::~SocketConnection () {

    turnBackgroundOff ();

    if (m_sockfd > 0) {
	close (m_sockfd);
    }
}

int SocketConnection::send (MESSAGE &message) 
{
    int code = MESSAGE_OK;
    MESSAGE *securedMessage = NULL;

	// make sure that we are in a connected state

    assert (Connection::getStatus() == Connection::e_CONNECTED);

	// encrypt the message if necessary

    if ((code = secureForServer (m_machine, m_security, message, securedMessage)) != MESSAGE_OK) {
	delete (securedMessage);
	return (code);
    }

	// send the messages

    if (securedMessage -> getFlag() != PROT_KEY) {

	code = message_send (*this, *securedMessage);

    } else {

	while (1) {		// something to break out of

	    int lengths[2];
	    char *buffers[2];
	   
	    if (securedMessage -> linearize() < 0) {
		code = MESSAGE_SYS_ERROR;
		break;
	    }

	    if (message.linearize() < 0) {
		code = MESSAGE_SYS_ERROR;
		break;
	    }

	    lengths[0] = securedMessage -> getSize ();
	    buffers[0] = securedMessage -> getBuffer ();
	    lengths[1] = message.getSize();
	    buffers[1] = message.getBuffer();

	    if (tcpip_writev (m_sockfd, 2, buffers, lengths) < (lengths[0] + lengths[1])) {
		code = MESSAGE_LOST_CONNECTION;
		break;
	    }

	    code = MESSAGE_OK;
	    break;
	}
    }

	/* done */

    delete (securedMessage);
    return (code);
}
	
int SocketConnection::unsecureSend (MESSAGE &message)
{
	// make sure that we are in a connected state

    assert (Connection::getStatus() == Connection::e_CONNECTED);

	// send the message

    return (message_send (*this, message));
}
 
MESSAGE * SocketConnection::unsecureReceive (const MESSAGE_SET &messageSet)
{
    MESSAGE *message; 

	// make sure that we are in a connected state

    assert (Connection::getStatus() == Connection::e_CONNECTED);

	// receive the message

    message = message_receive (*this, messageSet);
    return (message);
}

int SocketConnection::unsecureReceiveIntoBuffer (DynamicBuffer& buffer)
{   
    char *charBuffer;
    UINT_32 size;
    UINT_32 netSize;

	// make sure that we are in a connected state

    assert (Connection::getStatus() == Connection::e_CONNECTED);

	// read into the buffer

    if (SocketConnection::read (sizeof(UINT_32), (char *) &netSize) < (int) sizeof(UINT_32)) {
	return MESSAGE_CANT_READ;
    }

    size = ntohl (netSize);
    charBuffer = new char [size];

    if (SocketConnection::read (size, charBuffer) < ((int) size)) {
	delete (charBuffer);
	return MESSAGE_CANT_READ;
    }

    DynamicBuffer tempBuffer (size, charBuffer, BufferOwnership::e_TAKE);
    buffer = tempBuffer;
    return MESSAGE_OK;
}

int SocketConnection::unsecureSendFromBuffer (DynamicBuffer& buffer)
{
    int code;
    int size;
    UINT_32 netSize;

	// make sure that we are in a connected state

    assert (Connection::getStatus() == Connection::e_CONNECTED);

	// write from the buffer

    size = buffer.size ();
    netSize = htonl (size);

    if (SocketConnection::write (sizeof(UINT_32), (char *) &netSize) < (int) sizeof(UINT_32)) {
	code = MESSAGE_LOST_CONNECTION;
    } else if (SocketConnection::write (size, buffer.value()) < size) {
	code = MESSAGE_LOST_CONNECTION;
    } else {
	code = MESSAGE_OK;
    }

    return (code);
}

int SocketConnection::read (int n, char *buffer)
{
    int rc;

	// make sure that we are in a connected state

    assert (Connection::getStatus() == Connection::e_CONNECTED);

	// reset the socket error flag since we do not want "leftover" errors 
	// affecting the current operation - e.g., If we wrote to the socket  
	// while it was disconnected, we do not want to get an EPIPE error   
	// here; we want to read any available data and then get an EOF.      

    (void) tcpip_reset_error (m_sockfd);

	// read the data

    if ((rc = tcpip_readn (m_sockfd, buffer, n)) < n) {
	return MESSAGE_LOST_CONNECTION;
    }

    return (rc);
}

int SocketConnection::write (int n, const char *buffer)
{
    int rc;

	// make sure that we are in a connected state

    assert (Connection::getStatus() == Connection::e_CONNECTED);

	// write to the socket

    if ((rc = tcpip_writen (m_sockfd, buffer, n)) < 0) {
	return MESSAGE_LOST_CONNECTION;
    }

    return (rc);
}

void SocketConnection::sigioAlarm (int, int flags)
{
    Connection::ConnStatus status = getStatus();

	// have we just connected? 

    if ((status == Connection::e_WAITING_TO_CONNECT) && (flags & SigioTrap::e_WRITE_READY)) {
	setStatus (Connection::e_CONNECTED);
	m_trap -> connectionAlarm (*this, Connection::e_CONNECTED);
	m_handler -> changeFlags (SigioTrap::e_READ_READY);
	return;
    }

	// have we just gotten a connection request?

    if ((status == Connection::e_WAITING_TO_ACCEPT) && (flags & SigioTrap::e_READ_READY)) {
	setStatus (Connection::e_ACCEPT_READY);
	m_trap -> connectionAlarm (*this, Connection::e_ACCEPT_READY);
	return;
    }

	// are we reading from a connected socket?

    if ((status == Connection::e_CONNECTED) && (flags & SigioTrap::e_READ_READY)) {
	m_trap -> connectionAlarm (*this, Connection::e_READ_READY);
	return;
    }
	
	// ignore all other cases right now

    return;
}

void SocketConnection::turnBackgroundOn (ConnectionTrap *trap, MonitorGroup group)
{
    int flags;

	// assertions on the parameters

    assert (trap != NULL);
    assert (m_trap == NULL);
    assert (m_handler == NULL);

	// remember the trap since we are going to redirect through ours

    m_trap  = trap;

	// set up the handler

    flags = SigioTrap::e_READ_READY;

    if (getStatus() == Connection::e_WAITING_TO_CONNECT) {
	flags |= SigioTrap::e_WRITE_READY;
    }

    m_handler = new SigioHandler (m_sockfd, this, flags, group); 
    (void) tcpip_asynchronous (m_sockfd);
}

void SocketConnection::turnBackgroundOff (void)
{
    if (m_handler != NULL) {
	(void) tcpip_synchronous (m_sockfd);
	delete (m_handler);
	m_handler = NULL;
    }
}

int SocketConnection::checkReadyStatus (int /*inFlags*/, int &outFlags)
{
    int rc;
    struct timeval tv;
    register UINT_32 i;
    fd_mask readSet[MASK_SIZE];
    fd_mask writeSet[MASK_SIZE];

	/* default to nothing ready */

    outFlags = 0;

	/* use select if we are checking for both read and write */

    for (i = 0; i < MASK_SIZE; ++i) {
	readSet[i] = m_set[i];
	writeSet[i] = m_set[i];
    }

    do {
        tv.tv_sec  = 0;
        tv.tv_usec = 0;
	rc = select (m_sockfd + 1, (SELECT_MASK *) &readSet[0], (SELECT_MASK *) &writeSet[0], (SELECT_MASK *) NULL, &tv);
    } while ((rc < 0) && ((errno == EINTR) || (errno == EAGAIN)));

    if (rc < 0) {
	return MESSAGE_SYS_ERROR;
    }

	/* success so set the flags */

    if (readSet[m_index] > 0) {
	outFlags |= SigioTrap::e_READ_READY;
    } 
	
    if (writeSet[m_index] > 0) {
	outFlags |= SigioTrap::e_WRITE_READY;
    }

    return MESSAGE_OK;
}

TcpipClientConnection::TcpipClientConnection (const BASE_SECURITY &security):
    SocketConnection (security)
{
    // empty 
}

TcpipClientConnection::~TcpipClientConnection ()
{
    // empty
}

int TcpipClientConnection::connect (const MachineId &machine, UINT_32 port, struct timeval stop)
{
    int rc;

	// make sure that we are in the initialized state

    assert (Connection::getStatus() == Connection::e_INITIALIZED);

	// remember the machine and port

    m_machine = machine;
    m_port    = port;

	// mark that we are starting the connection

    Connection::setStatus (Connection::e_WAITING_TO_CONNECT);

	// get the IP address of the machine 

    if ((rc = tcpip_getip (m_machine, stop)) == TCPIP_TIMEOUT) {

	rc = MESSAGE_TIMEOUT;

    } else if (rc != TCPIP_OK) {

	rc = MESSAGE_NOT_FOUND;

    } else if ((m_sockfd = tcpip_socket ()) < 0) {

	rc = MESSAGE_SYS_ERROR;

    } else {

	FDMASK_ZERO (m_set);
	FDMASK_SET (m_sockfd, m_set);
	m_index = FDMASK_INDEX (m_sockfd);

        REMOTE_PORT serverInfo (m_machine.getName().value(), m_machine.getIp(), m_port);
        rc = tcpip_connect (m_sockfd, serverInfo, stop);

        switch (rc) {

	    case TCPIP_OK:	

	        rc = MESSAGE_OK;
	        break;

	    case TCPIP_NOT_FOUND:

	        rc = MESSAGE_NOT_FOUND;
	        break;

	    case TCPIP_TIMEOUT:

	        rc = MESSAGE_TIMEOUT;
	        break;

	    default:

	        rc = MESSAGE_NO_SERVER;
		break;
        }
    }

	// set the connection status 

    if (rc == MESSAGE_OK) {
	Connection::setStatus (Connection::e_CONNECTED);
    } else {
	Connection::setStatus (Connection::e_FAILED);
    }

	// done 

    return (rc);
}

int TcpipClientConnection::asyncConnect (const MachineId &machine, UINT_32 port)
{
    int rc;

	// assertions

    assert (machine.getIp() != UNKNOWN_IP);

	// make sure that we are in the initialized state

    assert (Connection::getStatus() == Connection::e_INITIALIZED);

	// remember the machine and port

    m_machine = machine;
    m_port    = port;

	// mark that we are starting the connection

    Connection::setStatus (Connection::e_WAITING_TO_CONNECT);

	// make the connection

    if ((m_sockfd = tcpip_socket ()) < 0) {

	rc = MESSAGE_SYS_ERROR;

    } else {

	FDMASK_ZERO (m_set);
	FDMASK_SET (m_sockfd, m_set);
	m_index = FDMASK_INDEX (m_sockfd);

        REMOTE_PORT serverInfo (m_machine.getName().value(), m_machine.getIp(), m_port);
        rc = tcpip_async_connect (m_sockfd, serverInfo);

        switch (rc) {

	    case TCPIP_OK:	

	        rc = MESSAGE_OK;
	        break;

	    case TCPIP_PENDING:

		rc = MESSAGE_PENDING;
		break;

	    default:

	        rc = MESSAGE_NOT_FOUND;
        }
    }

	// set the connection status 

    if (rc == MESSAGE_OK) {
	Connection::setStatus (Connection::e_CONNECTED);
    } else if (rc != MESSAGE_PENDING) {
	Connection::setStatus (Connection::e_FAILED);
    }

	// done 

    return (rc);
}

TcpipServerConnection::TcpipServerConnection (const BASE_SECURITY &security):
    SocketConnection (security)
{
    // empty 
}

TcpipServerConnection::~TcpipServerConnection ()
{
    // empty
}

UnixClientConnection::UnixClientConnection (const BASE_SECURITY &security):
    SocketConnection (security)
{
    // empty 
}

UnixClientConnection::~UnixClientConnection () {
    // empty
}

int UnixClientConnection::connect (const MachineId& machine, const DynamicString &socketFilename, struct timeval stop) 
{ 
    int rc; 

	// make sure that we are in the initialized state

    assert (Connection::getStatus() == Connection::e_INITIALIZED);

	// mark that we are starting the connection

    Connection::setStatus (Connection::e_WAITING_TO_CONNECT);

	// remember the machine and socket filename

    m_machine        = machine; 
    m_socketFilename = socketFilename; 

	// construct and connect the socket

    if ((m_sockfd = unix_socket ()) < 0) {

	rc = MESSAGE_SYS_ERROR;

    } else {

	FDMASK_ZERO (m_set);
	FDMASK_SET (m_sockfd, m_set);
	m_index = FDMASK_INDEX (m_sockfd);

        rc = unix_connect (m_sockfd, m_socketFilename.value(), stop);

        switch (rc) {

	    case TCPIP_OK:	

	        rc = MESSAGE_OK;
	        break;

	    case TCPIP_NOT_FOUND:

	        rc = MESSAGE_NOT_FOUND;
	        break;

	    case TCPIP_TIMEOUT:

	        rc = MESSAGE_TIMEOUT;
	        break;

	    default:

	        rc = MESSAGE_NO_SERVER;
	}
    }

	// set the connection status 

    if (rc == MESSAGE_OK) {
	Connection::setStatus (Connection::e_CONNECTED);
    } else {
	Connection::setStatus (Connection::e_FAILED);
    }

	// done

    return (rc);
}

UnixServerConnection::UnixServerConnection (const BASE_SECURITY &security):
    SocketConnection (security)
{
    // empty 
}

UnixServerConnection::~UnixServerConnection () {
    // empty
}

FileConnection::FileConnection (FileCallbacks &callbacks, const BASE_SECURITY &security):
    Connection (security),
    m_callbacks (callbacks)
{
    // empty
}

FileConnection::~FileConnection () {
    // empty
}

int FileConnection::checkReadyStatus (int inFlags, int &outFlags) 
{
    outFlags = inFlags;
    return MESSAGE_OK;
}

int FileConnection::read (int n, char *buffer)
{
    int rc;

    if ((rc = m_callbacks.read (n, buffer)) < 0) {
	return MESSAGE_CANT_READ;
    }

    return (rc);
}

int FileConnection::write (int n, const char *buffer)
{
    int rc;

    if ((rc = m_callbacks.write (n, buffer)) < 0) {
	return MESSAGE_CANT_WRITE;
    }

    return (rc);
}

int FileConnection::send (MESSAGE &message) 
{
    int code = MESSAGE_OK;
    MESSAGE *securedMessage = NULL;

	/* dummy machine for now */

    MachineId machine;

	/* encrypt and then send the message */

    if ((code = secureForServer (machine, m_security, message, securedMessage)) != MESSAGE_OK) {

	//empty

    } else if ((code = message_send (*this, *securedMessage)) != MESSAGE_OK) {

	// empty

    } else if (securedMessage -> getFlag() != PROT_KEY) {

	// empty

    } else {

	code = message_send (*this, message);
    }

	/* done */

    delete_check (securedMessage);
    return (code);
}

int FileConnection::unsecureSend (MESSAGE &message)
{
    int code = message_send (*this, message);
    return (code);
}

int FileConnection::unsecureSendFromBuffer (DynamicBuffer& buffer)
{
    int code;
    int size;
    UINT_32 netSize;

    size = buffer.size ();
    netSize = htonl (size);

    if (FileConnection::write (sizeof(UINT_32), (char *) &netSize) < (int) sizeof(UINT_32)) {
	code = MESSAGE_LOST_CONNECTION;
    } else if (FileConnection::write (size, buffer.value()) < size) {
	code = MESSAGE_LOST_CONNECTION;
    } else {
	code = MESSAGE_OK;
    }

    return (code);
}

MESSAGE * FileConnection::unsecureReceive (const MESSAGE_SET &messageSet)
{
    MESSAGE *response = message_receive (*this, messageSet);
    return (response);
}

int FileConnection::unsecureReceiveIntoBuffer (DynamicBuffer& buffer)
{
    char *charBuffer;
    UINT_32 size;
    UINT_32 netSize;

    if (FileConnection::read (sizeof(UINT_32), (char *) &netSize) < (int) sizeof(UINT_32)) {
	return MESSAGE_CANT_READ;
    }

    size = ntohl (netSize);
    charBuffer = new char [size];

    if (FileConnection::read (size, charBuffer) < ((int) size)) {
	return MESSAGE_CANT_READ;
    }

    DynamicBuffer tempBuffer (size, charBuffer, BufferOwnership::e_TAKE);
    buffer = tempBuffer;
    return MESSAGE_OK;
}
