/* Agent Tcl 
   Bob Gray
   26 June 1995

   genTcpip.cc

   This file implements the routines that handle TCP/IP connections.  These
   routines are based on the examples in [W. Richard Stevens, Unix Network
   Programming, Prentice Hall, 1990] and on the code that Saurab Nog
   developed as part of his A.I. project.

   Copyright (c) 1995, Robert S. Gray Dartmouth College

   See the file "agent.terms" for information on usage and redistribution
   of this file and for a DISCLAIMER OF ALL WARRANTIES.
*/

#ifndef NO_PRAGMAS
#pragma implementation
#endif

#include "platPorting.h"
#include "platSystem.h"
#include "platTimers.h"
#include "platTimeval.h"
#include "agentMachId.h"          // MachineId class
#include "genDString.h"

/* ------------------------------------------- */

#include <sys/un.h>
#include <setjmp.h>
#include "genTcpip.h"
#include "genUtility.h"
#include "truefalse.h"

    /* maximum name length for a Unix socket */

const MAX_NAME_LENGTH = 128;

    /* actual IP address and name of the current host */

static UINT_32 actual_ip   = 0;
static char *  actual_name = NULL;

    /* IP address "127.0.0.1" */

static UINT_32 local_ip = 0;

/* tcpip_setActual

   Purpose: Set the actual IP address and name of the current host.  Any 
            reference to the actual IP address or name will be turned into a 
            reference to "localhost".  This support is needed for mobile 
            computers that may have networking turned off.  If networking is 
            turned off, references to the actual IP address fail but 
	    references to "localhost" are fine.

     Input: ip   = actual IP address of the current host
		   (UINT_32)

	    name = name of the actual host
		   (char *)
 
    Output: The procedure sets the actual IP address and name of the current 
            host.
*/

void tcpip_setActual (UINT_32 ip, char *name)
{
    actual_ip   = ip;
    actual_name = my_strcpy (name); 
    local_ip    = (UINT_32) inet_addr ("127.0.0.1");
}

/* tcpip_getLocalhost

   Purpose: Return the numeric representation of the localhost IP address
            "127.0.0.1"

     Input: None

    Output: The procedure returns the numeric representaton of the localhost
            IP address "127.0.0.1".
*/
 
UINT_32 tcpip_getLocalhost (void) {
    
    if (local_ip == 0) {
	local_ip = (UINT_32) inet_addr ("127.0.0.1");
    }

    return (local_ip); 
}

/* tcpip_blocking

   Purpose: Turn on blocking I/O for a socket

     Input: sockfd = the socket descriptor

    Output: The procedure returns TCPIP_FCNTL on error and 0 otherwise.
*/

int tcpip_blocking (int sockfd)
{
    int mode;
    int newMode;

    if ((mode = fcntl (sockfd, F_GETFL, 0)) < 0) {
	return TCPIP_FCNTL;
    }

    newMode = mode & ~NON_BLOCKING_FLAG;

    if (fcntl (sockfd, F_SETFL, newMode) < 0) {
	return TCPIP_FCNTL;
    }

    return TCPIP_OK;
}

/* tcpip_nonblocking

   Purpose: Turn on nonblocking I/O for a socket

     Input: sockfd = the socket descriptor

    Output: The procedure returns TCPIP_FCNTL on error and 0 otherweise.
*/

int tcpip_nonblocking (int sockfd)
{
    int mode;
    int newMode;

    if ((mode = fcntl (sockfd, F_GETFL, 0)) < 0) {
	return TCPIP_FCNTL;
    }

    newMode = mode | NON_BLOCKING_FLAG;

    if (fcntl (sockfd, F_SETFL, newMode) < 0) {
	return TCPIP_FCNTL;
    }

    return TCPIP_OK;
}

/* tcpip_asynchronous

   Purpose: Turn on asynchronous I/O for a socket

     Input: sockfd = the socket descriptor

    Output: The procedure returns TCPIP_FCNTL on error and 0 otherwise.
*/

int tcpip_asynchronous (int sockfd)
{ 
  int flags;

  if (fcntl (sockfd, F_SETOWN, getpid()) < 0) {
    return TCPIP_FCNTL;
  }

  if ((flags = fcntl (sockfd, F_GETFL, 0)) < 0) {
    return TCPIP_FCNTL;
  }

  if (fcntl (sockfd, F_SETFL, flags | FASYNC) < 0) {
    return TCPIP_FCNTL;
  }

  return 0;
}

/* tcpip_synchronous

   Purpose: Turn off asynchornous I/O for a socket

     Input: sockfd = the socket descriptor

    Output: The procedure returns TCPIP_FCNTL on error and 0 otherwise.
*/

int tcpip_synchronous (int sockfd)
{ 
    int flags;

    if ((flags = fcntl (sockfd, F_GETFL, 0)) < 0) {
	return TCPIP_FCNTL;
    }

    if (fcntl (sockfd, F_SETFL, flags & ~FASYNC) < 0) {
	return TCPIP_FCNTL;
    }

    return 0;
}

/* tcpip_getsockname

   Purpose: Get the IP address and the port number associated with the
            LOCAL end of a socket

     Input: sockfd = socket descriptor

    Output: The procedure returns TCPIP_UNABLE_TO_GET_PORT on error.
            Otherwise the procedure fills in *ip and *port with the IP 
            address and port number.
*/

int tcpip_getsockname (int sockfd, UINT_32 *ip, UINT_32 *port)
{
  struct sockaddr_in address;  
  int size = sizeof(sockaddr_in);

  if (getsockname(sockfd, (struct sockaddr *) &address, &size) < 0) {
    return TCPIP_UNABLE_TO_GET_PORT;
  }

  *ip   = (UINT_32) address.sin_addr.s_addr;
  *port = ntohs (address.sin_port);
  return 0;
}

/* tcpip_getpeername

   Purpose: Get the IP address and the port number associated with the
            REMOTE end of a CONNECTED socket

     Input: sockfd = socket descriptor

    Output: The procedure returns TCPIP_UNABLE_TO_GET_PORT on error.
            Otherwise the procedure fills in *ip and *port with the IP 
            address and port number.
*/

int tcpip_getpeername (int sockfd, UINT_32 *ip, UINT_32 *port)
{
  struct sockaddr_in address;  
  int size = sizeof(sockaddr_in);

  if (getpeername(sockfd, (struct sockaddr *) &address, &size) < 0) {
    return TCPIP_UNABLE_TO_GET_PORT;
  }

  *ip   = (UINT_32) address.sin_addr.s_addr;  
  *port = ntohs (address.sin_port);
  return 0;
}

/* tcpip_getip

   Purpose: This procedure gets the IP address of the given machine *but* 
            failes with an error if it is unable to get the IP address before
            the specified stop time.  This procedure represents an *evil* use
            of the "sigsetjmp" and "siglongjmp" standard C functions but there
            is no choice since "gethostbyname" will not abort until its 
            internal timer expires; a SIGALRM set by the program will
            interrupt "gethostbyname" but "gethostbyname" will always resume
            when the SIGALRM handler has finished.  Since the internal timer
            of "gethostbyname" ranges from 2 to 5 minutes on many systems,
            this is unacceptable.  Therefore I have resorted to calling
            "siglongjmp" from inside the SIGALRM handler.  If I can ever find a
            better solution, I will use it immediately.  

#ifdef FIX_LATER

	    Problem: "siglongjmp", of course, is not an async-safe function 
            which means that we are taking a risk here.  We might interrupt 
            gethostbyname while it is doing some memory allocation.  We also 
            might interrupt gethostbyname while it has unfreed, temporary 
            memory allocated, leading to a memory leak.  So far, however, 
            running the risk seems to be worth it.

	    Update: 02/05/98: Now that I am within days of the next release,
	    I have decided that the siglongjmp is too risky.  Although it has 
            never given us any problems, it seems foolish to release a program
            with a known, potential memory corruption problem.  So I have
            removed all the time code from this routine by commenting it out
            with "#ifdef RISKY_USE_OF_SIGLONGJMP".

	    Question: Does anyone out there know how I can safely impose a
            timeout on gethostbyname in a single-threaded environment?

	    Problem: Also this arrangement is completely unacceptable for a
            multi-threaded program.  Although I am using this function in the
            Java and Python modules still, it is only because we can 
            easily test most of the system in such a way that a timeout never
            occurs here.  We must move tcpip_getip down into the platform layer
            and provide alternative implementations for the multi-threaded
            environments.

#endif


     Input: machine = machine name and IP address
		      (class MachineId &) 

	    stop    = wall time at which to give up
		      (struct timeval)
 
    Output: The procedure returns one of the following codes.

	    TCPIP_TIMEOUT	= timeout expired before IP address was found
	    TCPIP_NOT_FOUND	= unable to determine IP address
	    TCPIP_OK		= success

	    On success machine.ip is filled in with the IP address of the
	    machine.     
*/

#ifdef RISKY_USE_OF_SIGLONGJMP

class HostnameTrap: public TimerTrap
{
	Timers timers;            // timers
	sig_atomic_t inProgress;   // getip in progress

    public:

	sigjmp_buf buffer;        // buffer for siglongjmp

	HostnameTrap (void):
	    inProgress (0)
	{}

        void wallAlarm (void) {

	    if (!inProgress) {
		delete (this);
	    } else {
		siglongjmp (buffer, 1);
	    }
	}

	void cpuAlarm (void) {
	    abort_with_message ("HostnameTrap::cupAlarm should never get called!");
	}

	void start (struct timeval stop) {
	    inProgress = 1;
	    timers.add (stop, this);
	}

	void stop (void) {
	    inProgress = 0;
	}

	void dispose (void) {
	    inProgress = 0;
	    timers.disposeTimers ();
	}
};

static HostnameTrap *s_timerTrap = NULL;

#endif

static char         *s_name      = NULL;

int tcpip_getip (MachineId& machine, struct timeval stop)
{
    int rc = TCPIP_OK;
    int length;
    UINT_32 ip;
    struct hostent *serverInfo;

	/* assertions on the parameters */

    assert ((machine.getName().isEmpty() == e_FALSE) || (machine.getIp() != UNKNOWN_IP));
    assert ((stop.tv_sec >= 0) && (stop.tv_usec >= 0));

	/* do nothing if we know the IP address */

    if (machine.getIp() != UNKNOWN_IP) {
	return TCPIP_OK;
    }

	/* see if we have something to get the IP address of */

    if (((s_name = strip_whitespace (machine.getName().value())) != NULL) && (*s_name != '\0')) { 

	    /* see if the name corresponds to the current host */

	if (actual_name != NULL) {

	    length = strlen (s_name);

	    if (!strncmp (s_name, actual_name, length)) {
		if ((s_name[length-1] == '.') || (actual_name[length] == '.') || (actual_name[length] == '\0')) {
		    machine.setIp (actual_ip);
		    delete (s_name);
		    return TCPIP_OK;
		}
	    }
	}

	    /* see if the name is an ASCII represention of an IP */ 

	if ((ip = (UINT_32) inet_addr (s_name)) != INADDR_NONE) {
	    machine.setIp (ip);
	    delete (s_name);
	    return (TCPIP_OK);
	}

#ifdef RISKY_USE_OF_SIGLONGJMP

	    /* otherwise go to the name server; first set up the timer trap */

	s_timerTrap = new HostnameTrap ();

	    /* jump back here when the timer trap is called; make sure */
	    /* that we call disposeTimers since otherwise we will miss */
	    /* timers; then bail with a timeout error                  */

	if (sigsetjmp (s_timerTrap -> buffer, 1) == 1) {
	    s_timerTrap -> dispose ();
	    delete (s_timerTrap);
	    delete (s_name);
	    return TCPIP_TIMEOUT; 
        } 

	    /* gethostbyname with a timer */ 

        s_timerTrap -> start (stop);

#endif

	serverInfo = gethostbyname (s_name);

#ifdef RISK_USE_OF_SIGLONGJMP

	s_timerTrap -> stop ();

#endif

	    /* did we get the IP address */

	if (serverInfo != NULL) {

	    memcpy ((void *) &ip, (void *) serverInfo -> h_addr, sizeof(UINT_32));
	    machine.setIp (ip);
	    rc = TCPIP_OK;

	} else {

	    rc = TCPIP_NOT_FOUND;
	}
    }

	/* done -- do NOT delete the timer trap since the timer might   */
	/* still be active -- it would be nice to have a "cancel timer" */
	/* facility here (we do have a canel timer facility now)        */

    delete (s_name);
    return (rc);
}
      
/* tcpip_hostname

   Purpose: Get the name of the current host

     Input: None

    Output: The procedure returns NULL if it is unable to get the host name.
            Otherwise the procedure returns a pointer to a dyanmically
            allocated string that contains the hostname.
*/

char *tcpip_hostname (void)
{
    char *nameCopy;
    char name[MAX_NAME_LENGTH];

    if (gethostname (name, MAX_NAME_LENGTH) < 0) {
	return NULL;
    }

    nameCopy = new char [strlen(name) + 1];
    strcpy (nameCopy, name);

    for (char *sp = nameCopy; *sp != '\0'; ++sp) {
	if ((*sp >= 'A') && (*sp <= 'Z')) {
	    *sp = *sp - 'A' + 'a';
	}
    }

    return (nameCopy);
}

/* tcpip_socket

   Purpose: This function creates a socket.

     Input: None

   Purpose: The procedure returns TCPIP_UNABLE_TO_CREATE if it is unable to
            create the socket.  Otherwise the procedure returns the sockfd.
*/

int tcpip_socket (void)
{
    int sockfd;
    int flag = 1;

	/* create the socket */

    if ((sockfd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
	return TCPIP_UNABLE_TO_CREATE;
    }

	/* turn on the TCP_NODELAY flag */

    if (setsockopt (sockfd, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int)) < 0) {
	close (sockfd);
	return TCPIP_UNABLE_TO_CREATE;
    }

    return sockfd;
}

/* unix_socket

   Purpose: This function creates a Unix domain socket.

     Input: None

    Output: The procedure returns TCPIP_UNABLE_TO_CREATE if it is unable to
            create the socket.  Otherwise the procedure returns the sockfd.
*/

int unix_socket (void)
{
    int sockfd;

    if ((sockfd = socket (AF_UNIX, SOCK_STREAM, 0)) < 0) {
	return TCPIP_UNABLE_TO_CREATE;
    }

    return sockfd;
}
 
/* tcpip_async_connect 

   Purpose: Connect a socket to a port on a remote machine wihout waiting
	    for the connection to complete here

     Input: sockfd    = socket descriptor
		        (int)

	    port      = machine name and IP address and the port number
			(struct REMOTE_PORT &).  IP address must be known!

    Output: The procedure returns one of the following codes.

            TCPIP_UNABLE_TO_CONNECT = connection request was rejected
	    TCPIP_PENDING	    = connection is completing
	    TCPIP_OK                = success
*/

int tcpip_async_connect (int sockfd, REMOTE_PORT& port)
{
    int code;
    struct sockaddr_in serverAddress;    

	/* assertions */

    assert (port.ip != UNKNOWN_IP);

	/* use "localhost" for the local host */

    if (port.ip == actual_ip) {
	port.ip = local_ip;
    }

	/* fill in the server address */

    memset ((void *) &serverAddress, 0, sizeof (sockaddr_in));
    memcpy ((void *) &serverAddress.sin_addr, (void *) &port.ip, sizeof(UINT_32));
    serverAddress.sin_family = AF_INET;
    serverAddress.sin_port   = htons (port.port);

	/* mark the socket as nonblocking */

    (void) tcpip_nonblocking (sockfd);
    (void) tcpip_reset_error (sockfd);

	/* start the asynchronous connection */

    do {
	code = connect (sockfd, (struct sockaddr *) &serverAddress, sizeof(sockaddr_in));
    } while ((code < 0) && ((errno == EINTR) || (errno == EAGAIN)));

	/* see if the connection has completed already */

    if (code == 0) {
	code = TCPIP_OK;
    } else if ((errno == EWOULDBLOCK) || (errno == EINPROGRESS)) {
	code = TCPIP_PENDING;
    } else {
	code = TCPIP_UNABLE_TO_CONNECT;
    }

	/* mark the socket as blocking */

    (void) tcpip_blocking (sockfd);
    return (code);
}

/* tcpip_connect 

   Purpose: Connect a socket to a port on a remote machine

     Input: sockfd    = socket descriptor
		        (int)

	    port      = machine name and IP address and the port number
			(struct REMOTE_PORT &)

	    stop      = wall time at which to abandon the connection attempt
		        (struct timeval)
						
    Output: The procedure returns one of the following codes.

            TCPIP_NOT_FOUND         = unable to determine network location of destination machine
	    TCPIP_TIMEOUT           = timeout expired before connection
            TCPIP_UNABLE_TO_CONNECT = connection request was rejected
	    TCPIP_OK                = success
*/

int tcpip_connect (int sockfd, REMOTE_PORT& port, struct timeval stop)
{
    int code;
    int serrno;
    int serrnoLen = sizeof(int);
    struct timeval timeout; 
    fd_mask writeSet[MASK_SIZE];
    struct sockaddr_in serverAddress;    

	/* convert the machine name to an IP number */

    MachineId machine (port.machine, port.ip);

    if ((code = tcpip_getip (machine, stop)) != TCPIP_OK) { 
	return (code);
    }

	/* use "localhost" for the local host */

    if (machine.getIp() == actual_ip) {
	machine.setIp (local_ip);
    }

	/* fill in the server address */

    UINT_32 ip = machine.getIp();
    memset ((void *) &serverAddress, 0, sizeof (sockaddr_in));
    memcpy ((void *) &serverAddress.sin_addr, (void *) &ip, sizeof(UINT_32));
    serverAddress.sin_family = AF_INET;
    serverAddress.sin_port   = htons (port.port);

	/* mark the socket as nonblocking */

    (void) tcpip_nonblocking (sockfd);
    (void) tcpip_reset_error (sockfd);

	/* start the asynchronous connection */

    do {
	code = connect (sockfd, (struct sockaddr *) &serverAddress, sizeof(sockaddr_in));
    } while ((code < 0) && ((errno == EINTR) || (errno == EAGAIN)));

	/* wait for the connection to complete */

    if (code < 0) {

	if ((errno == EWOULDBLOCK) || (errno == EINPROGRESS)) {

	    do {

		FDMASK_ZERO (writeSet);
		FDMASK_SET (sockfd, writeSet);

		timeout = TimevalUtil::subTimevals (stop, SystemUtil::getCurrentWall());

		code = select (sockfd + 1, (SELECT_MASK *) NULL, (SELECT_MASK *) &writeSet[0], (SELECT_MASK *) NULL, &timeout);

	    } while ((code < 0) && ((errno == EINTR) || (errno == EAGAIN)));
	}

	if (code < 0) {
	    code = TCPIP_UNABLE_TO_CONNECT;
	} else if (code == 0) {
	    code = TCPIP_TIMEOUT;
	} else if (getsockopt (sockfd, SOL_SOCKET, SO_ERROR, (char *) &serrno, &serrnoLen) < 0) {
	    code = TCPIP_UNABLE_TO_CONNECT;
	} else if (serrno != 0) {
	    code = TCPIP_UNABLE_TO_CONNECT;
	} else {
	    code = TCPIP_OK;
	}
    }

	/* undo the mode change */

    (void) tcpip_blocking (sockfd);


	/* turn on the TCP_NODELAY flag */

    int flag = 1;

    if (setsockopt (sockfd, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int)) < 0) {
	close (sockfd);
	return TCPIP_UNABLE_TO_CREATE;
    }

    return (code);
}

/* unix_connect

   Purpose: Connect to a Unix domain socket

     Input: sockfd = socket descriptor
            name   = filename of the Unix domain socket

    Output: The procedure returns TCPIP_UNABLE_TO_CONNECT if it is unable to
	    connect sockfd to the Unix domain socket.  Otherwise the procedure
	    returns 0.
*/

int unix_connect (int sockfd, char *name, struct timeval stop)
{
    int code;
    int serrno;
    int serrnoLen = sizeof(int);
    unsigned len;
    struct timeval timeout; 
    fd_mask writeSet[MASK_SIZE];
    struct sockaddr_un unix_addr;

	/* determine the appropriate length */

#ifdef SUNLEN_EXISTS
    len = sizeof(unix_addr.sun_family) + sizeof(unix_addr.sun_len) + strlen(name) + 1;
#else
    len = sizeof(unix_addr.sun_family) + strlen(name);
#endif

	/* check the length */

    if (len >= sizeof(sockaddr_un)) {
	return TCPIP_UNABLE_TO_CONNECT;
    }

	/* fill in the sockaddr structure */

    memset ((void *) &unix_addr, 0, sizeof(sockaddr_un));
#ifdef SUNLEN_EXISTS
    unix_addr.sun_len = len;
#endif
    unix_addr.sun_family = AF_UNIX;
    strcpy (unix_addr.sun_path, name);

	/* mark the socket as nonblocking */

    (void) tcpip_nonblocking (sockfd);
    (void) tcpip_reset_error (sockfd);

	/* start the asynchronous connection */

    do {
	code = connect (sockfd, (struct sockaddr *) &unix_addr, len);
    } while ((code < 0) && ((errno == EINTR) || (errno == EAGAIN)));

	/* wait for the connection to complete */

    if (code < 0) {

	if ((errno == EWOULDBLOCK) || (errno == EINPROGRESS)) {

	    do {

		FDMASK_ZERO (writeSet);
		FDMASK_SET (sockfd, writeSet);

		timeout = TimevalUtil::subTimevals (stop, SystemUtil::getCurrentWall());

		code = select (sockfd + 1, (SELECT_MASK *) NULL, (SELECT_MASK *) &writeSet[0], (SELECT_MASK *) NULL, &timeout);

	    } while ((code < 0) && ((errno == EINTR) || (errno == EAGAIN)));
	}

	if (code < 0) {
	    code = TCPIP_UNABLE_TO_CONNECT;
	} else if (code == 0) {
	    code = TCPIP_TIMEOUT;
	} else if (getsockopt (sockfd, SOL_SOCKET, SO_ERROR, (char *) &serrno, &serrnoLen) < 0) {
	    code = TCPIP_UNABLE_TO_CONNECT;
	} else if (serrno != 0) {
	    code = TCPIP_UNABLE_TO_CONNECT;
	} else {
	    code = TCPIP_OK;
	}
    }

	/* mark the socket as blocking */

    (void) tcpip_blocking (sockfd);
    return (code);
}
 
int unix_async_connect (int sockfd, char *name)
{
    int code;
    unsigned len;
    struct sockaddr_un unix_addr;

	/* determine the appropriate length */

#ifdef SUNLEN_EXISTS
    len = sizeof(unix_addr.sun_family) + sizeof(unix_addr.sun_len) + strlen(name) + 1;
#else
    len = sizeof(unix_addr.sun_family) + strlen(name);
#endif

	/* check the length */

    if (len >= sizeof(sockaddr_un)) {
	return TCPIP_UNABLE_TO_CONNECT;
    }

	/* fill in the sockaddr structure */

    memset ((void *) &unix_addr, 0, sizeof(sockaddr_un));
#ifdef SUNLEN_EXISTS
    unix_addr.sun_len = len;
#endif
    unix_addr.sun_family = AF_UNIX;
    strcpy (unix_addr.sun_path, name);

	/* mark the socket as nonblocking */

    (void) tcpip_nonblocking (sockfd);
    (void) tcpip_reset_error (sockfd);

	/* start the asynchronous connection */

    do {
	code = connect (sockfd, (struct sockaddr *) &unix_addr, len);
    } while ((code < 0) && ((errno == EINTR) || (errno == EAGAIN)));

	/* see if the connection has completed already */ 

    if (code == 0) {
	code = TCPIP_OK;
    } else if ((errno == EWOULDBLOCK) || (errno == EINPROGRESS)) {
	code = TCPIP_PENDING;
    } else {
	code = TCPIP_UNABLE_TO_CONNECT;
    }

	/* mark the socket as blocking */

    (void) tcpip_blocking (sockfd);
    return (code);
}

/* tcpip_listen

   Purpose: Listen to a socket

     Input: sockfd = socket descriptor

    Output: The procedure returns TCPIP_UNABLE_TO_LISTEN on error.  Otherwise
            the procedure returns 0.
*/

int tcpip_listen (int sockfd)
{
  if (listen(sockfd, TCPIP_BACKLOG) < 0)
  {
    return TCPIP_UNABLE_TO_LISTEN;
  }
  else
  {
    return 0;
  }
}

/* tcpip_bind

   Purpose: Bind a socket to a local port

     Input: sockfd = socket descriptor
            port   = port number

    Output: The procedure returns TCPIP_UNABLE_TO_BIND if it can not bind to
            the port.  The procedure returns TCPIP_UNABLE_TO_GET_PORT if it
	    is unable to verify the port number of the bound port.  Otherwise
	    the procedure returns 0 and fills in *selected_port with the 
	    TCPIP/IP port of the socket.  The selected port will be equal to 
	    the requested port unless the requested port was TCPIP_ANY in 
	    which case the procedure lets the system choose a port.
*/

const int c_MAXIMUM_BIND_FAILURES = 5;

int tcpip_bind (int sockfd, int port, int *selected_port)
{
    int bindFailures = 0;
    struct sockaddr_in address;
    int size = sizeof(sockaddr_in);

	/* keep trying if we are asking the system for a port */

    while (1) {

	    /* bind the local address so thet a client can send to us */

	memset ((void *) &address, 0, sizeof(address));
	address.sin_family      = AF_INET;
	address.sin_port        = htons (port);
	address.sin_addr.s_addr = htonl (INADDR_ANY);

	if (bind(sockfd, (struct sockaddr *) &address, size) < 0) {

	    if ((port == TCPIP_ANY_PORT) && (errno == EADDRINUSE)) {

		if (++bindFailures < c_MAXIMUM_BIND_FAILURES) {
		    printf ("BIND FAILURE ==> %d\n", bindFailures);
		    continue;
		}

		abort_with_message ("tcpip_bind: got multiple EADDRINUSE errors even though OS is selecting the port number");
	    }

	    return (TCPIP_UNABLE_TO_BIND);
	}

	break;
    }

    if (getsockname(sockfd, (struct sockaddr *) &address, &size) < 0) {
	return (TCPIP_UNABLE_TO_GET_PORT);
    }

    if (selected_port != NULL) {
	*selected_port = ntohs (address.sin_port);
    }
 
    return (TCPIP_OK);
}

/* unix_bind

   Purpose: Bind a Unix domain socket to a filename

     Input: sockfd = socket descriptor
            name   = filename  

    Output: The procedure returns TCPIP_UNABLE_TO_BIND if it is unable to
            bind the socket to the specified port.  Otherwise the procedure
            returns 0.
*/

int unix_bind (int sockfd, char *name)
{
  unsigned len;
  struct sockaddr_un unix_addr;

    /* determine the appropriate length */

#ifdef SUNLEN_EXISTS
  len = sizeof(unix_addr.sun_family) + sizeof(unix_addr.sun_len) + strlen(name) + 1;
#else
  len = sizeof(unix_addr.sun_family) + strlen(name);
#endif

    /* check the length */

  if (len >= sizeof(unix_addr)) {
    return TCPIP_UNABLE_TO_BIND;
  }

    /* remove the filename if it exists */

  unlink (name);

    /* create the sockaddr structure */

  memset ((void *) &unix_addr, 0, sizeof(unix_addr));
#ifdef SUNLEN_EXISTS
  unix_addr.sun_len = len;
#endif
  unix_addr.sun_family = AF_UNIX;
  strcpy (unix_addr.sun_path, name);

    /* bind to the filename */

  if (bind (sockfd, (struct sockaddr *) &unix_addr, len) < 0) {
    return TCPIP_UNABLE_TO_BIND;
  }

  return 0;
}

/* tcpip_setup

   Purpose: This function creates, binds and listens to a socket.

     Input: port = TCP/IP port for the socket

    Output: The procedure returns one of the error flags in tcpip.h on error.
            Otherwise the procedure fills in *selected__port with the TCP/IP
            port of the socket and returns the socket descriptor of the 
            socket.  The selected port will be equal to the requested port
            unless the desired port is TCPIP_ANY in which case the procedure 
            lets the system choose a port.
*/

int tcpip_setup (int port, int *selected_port)
{
  int sockfd;                  /* socket descriptor   */
  int return_code;             /* return code         */

    /* open the socket */

  if ((sockfd = tcpip_socket ()) < 0) {
    return sockfd;
  }

    /* bind the socket */

  if ((return_code = tcpip_bind (sockfd, port, selected_port)) < 0) {
    close (sockfd);
    return return_code;
  }

    /* listen at the sockt */

  if ((return_code = tcpip_listen (sockfd)) < 0) {
    close (sockfd);
    return return_code;
  }

  return sockfd;
}

/* unix_setup

   Purpose: This function creates, binds and listens to a Unix domain socket.

     Input: name = filename

    Output: The procedure returns one of the error flags in tcpip.h on error.
            Otherwise the procedure returns the socket number.
*/

int unix_setup (char *name)
{
  int sockfd;                  /* socket descriptor   */
  int return_code;             /* return code         */

    /* open the socket */

  if ((sockfd = unix_socket ()) < 0) {
    return sockfd;
  }

    /* bind the socket */

  if ((return_code = unix_bind (sockfd, name)) < 0) {
    close (sockfd);
    return return_code;
  }

    /* listen at the sockt */

  if ((return_code = tcpip_listen (sockfd)) < 0) {
    close (sockfd);
    return return_code;
  }

  return sockfd;
}

/* tcpip_accept 

   Purpose: Accept a connection on a socket

     Input: sockfd = socket that has been bound to a port and that has been 
		     put into listening mode -- i.e., a socket that has been
		     returned from tcpip_setup 
		     (int)

	    stop   = time at which to give up
		     (struct timeval)
 
    Output: The procedure returns the socket descriptor of the new accepted
	    connection on success.  Otherwise the procedure returns one of
	    the following error codes:

 
 
*/

int tcpip_accept (int sockfd, struct timeval stop)
{
    int rc;				// return code from this procedure
    int code;				// return code from "select"
    int flag = 1;                    	// flag for setsockopt        
    int newSockfd;                  	// new socket descriptor      
    struct timeval timeout;		// timeout for "select"
    fd_mask readSet[MASK_SIZE];		// read mask for "select"
    struct sockaddr_in address;      	// client address           
    int size = sizeof(sockaddr_in);  	// size of the client address

	// put the socket into nonblocking mode */

    tcpip_nonblocking (sockfd);
    tcpip_reset_error (sockfd);

	// loop until acceptance or timeout

    while (1) {

	    // do a nonblocking accept

	if ((newSockfd = accept (sockfd, (struct sockaddr *) &address, &size)) >= 0) {

		// make sure the TCP_NODELAY option is set for the new socket     
		// This should not be necessary but on some machines the accept()
		// function does not correctly propagate the TCP_NODELAY flag   
		// from sockfd to newSockfd.                                     

	    setsockopt (newSockfd, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int));
	    rc = TCPIP_OK;
	    break;
	}

	    // any code other than EAGAIN indicates a fatal socket error

	if ((errno != EAGAIN) && (errno != EWOULDBLOCK)) {
	    rc = TCPIP_UNABLE_TO_ACCEPT;
	    break;
	}

	    // wait for a connection request

	do {

	    FDMASK_ZERO (readSet);
	    FDMASK_SET (sockfd, readSet);

	    timeout = TimevalUtil::subTimevals (stop, SystemUtil::getCurrentWall());

	    code = select (sockfd + 1, (SELECT_MASK *) &readSet[0], (SELECT_MASK *) NULL, (SELECT_MASK *) NULL, &timeout);

	} while ((code < 0) && ((errno == EINTR) || (errno == EAGAIN)));

	    // check for fatal socket error or timeout

	if (code < 0) {
	    rc = TCPIP_UNABLE_TO_ACCEPT;
	    break;
	} 

	if (code == 0) {
	    rc = TCPIP_TIMEOUT;
	    break;
	}
    }

	// mark the socket as blocking and return

    tcpip_blocking (sockfd);
    return ((rc == TCPIP_OK) ? newSockfd : rc);
}

/* tcpip_readn

   Purpose: This function is used in place of "read" when reading from a
            socket.

     Input: fd     = socket descriptor
            ptr    = buffer into which the characters are read
            nbytes = number of characters to read

    Output: The procedure returns an integer less than 0 on error.  
            Otherwise the procedure returns the number of characters
            read from the socket.
*/

int tcpip_readn (int fd, char *ptr, size_t nbytes)
{
  register int nread;               /* number of bytes read         */   
  register size_t nleft = nbytes;   /* number of bytes left to read */

  while (nleft > 0) {

    do {
      nread = read (fd, ptr, nleft);
    } while ((nread < 0) && (((errno == EINTR) || (errno == EAGAIN)) || (errno == EAGAIN)));

    if (nread < 0) {             /* ERROR */
      return (nread);
    } else if (nread == 0) {     /* EOF   */
      break;            
    }
 
    nleft -= nread;
    ptr   += nread;
  }

  return (nbytes - nleft);
}

/* tcpip_writev

   Purpose: This function is used in place of "writev" when writing to a
            socket.

     Input: fd          = file descriptor
		          (int)

	    bufferCount = number of buffers
			  (int)

	    buffers     = the buffers
			  (char **)

	    lengths     = the buffer lengths
			  (int *)

    Output: The procedure returns -1 on error.
            The procedure returns 0 on EOF. 
            Otherwise the procedure returns the number of bytes written.
*/

const int STATIC_BUFFER_SIZE	= 1536;
	    
int tcpip_writev (int fd, int bufferCount, char **buffers, int *lengths)
{
#ifdef NO_WRITEV_FUNCTION 

    int i;
    int totalLength = 0;

    for (i = 0; i < bufferCount; ++i) {
	totalLength += lengths[i];
    }

    char *buffer = new char [totalLength];
    char *bp     = buffer;
    
    for (i = 0; i < bufferCount; ++i) {
	memcpy (bp, buffers[i], lengths[i]);
	bp += lengths[i];
    }

    return (tcpip_writen (fd, buffer, totalLength));

#else

    int i;
    char *bp;
    int nleft;
    int nwritten;
    int buffersLeft;
    int totalLength = 0;
    struct iovec *ioPtr;
    struct iovec *ioBuffers = new struct iovec [bufferCount];
    char staticBuffer[STATIC_BUFFER_SIZE];

    for (i = 0; i < bufferCount; ++i) {
	totalLength += lengths[i];
    }

    if (totalLength < STATIC_BUFFER_SIZE) {

	bp = staticBuffer;

	for (i = 0; i < bufferCount; ++i) {
	    memcpy (bp, buffers[i], lengths[i]);
	    bp += lengths[i];
	}

	return (tcpip_writen (fd, staticBuffer, totalLength));
    }

    for (i = 0; i < bufferCount; ++i) {
	ioBuffers[i].iov_base = buffers[i];
	ioBuffers[i].iov_len  = lengths[i];
    }

    nleft       = totalLength;
    buffersLeft = bufferCount;
    ioPtr       = ioBuffers;

    while (nleft > 0) {

	do {
	    nwritten = writev (fd, ioPtr, buffersLeft);
	} while ((nwritten < 0) && ((errno == EINTR) || (errno == EAGAIN)));

	if (nwritten <= 0) {
	    break;
	}

	nleft -= nwritten;

	if (nleft > 0) {

	    while (nwritten > ioPtr -> iov_len) {
		nwritten -= ioPtr -> iov_len;
		buffersLeft -= 1;
		ioPtr += 1;
	    }

	    ioPtr -> iov_len -= nwritten;
	}
    }

    delete (ioBuffers);
    return (totalLength - nleft);

#endif
}

/* tcpip_writen

   Purpose: This function is used in place of "write" when writing to a 
            socket.

     Input: fd     = file descriptor
		     (int)

            ptr    = buffer from which characters are written
		     (char *)

            nbytes = number of bytes to write
		     (int)

    Output: The procedure returns -1 on error.
            The procedure returns 0 on EOF. 
            Otherwise the procedure returns the number of bytes written.
*/

int tcpip_writen (int fd, const char *ptr, size_t nbytes)
{
   register size_t nleft = nbytes;   /* number of bytes left to write  */
   register int nwritten;            /* number of bytes written        */
   
   nleft = nbytes;

   while (nleft > 0)
   {
     do
     {
       nwritten = write (fd, ptr, nleft);
     }
     while ((nwritten < 0) && ((errno == EINTR) || (errno == EAGAIN)));

     if (nwritten <= 0) {         /* ERROR or EOF */	
        return nwritten;
     }

     nleft -= nwritten;
     ptr   += nwritten;
   }

   return (nbytes - nleft);
}

/* tcpip_error_to_string

   Purpose: Convert one of the error codes to a string

     Input: code = error code

    Output: The procedure returns a pointer to a STATIC string that describes
            the error code. 
*/

const char *tcpip_error_to_string (int code)
{
   switch (code)
   {
     case TCPIP_UNABLE_TO_CREATE:
       return "unable to open a socket";
     case TCPIP_UNABLE_TO_BIND:
       return "unable to bind sockfd to the port";
     case TCPIP_UNABLE_TO_LISTEN:
       return "unable to listen on sockfd "; 
     case TCPIP_UNABLE_TO_GET_PORT:
       return "unable to get the port number of sockfd";
     case TCPIP_UNABLE_TO_ACCEPT:
       return "unable to accept on sockfd";
     case TCPIP_NOT_FOUND:
       return "unable to determine network address of machine"; 
     case TCPIP_UNABLE_TO_CONNECT:
       return "unable to connect sockfd to remote port";
     case TCPIP_TIMEOUT:
       return "timeout";
     default:
       return "unknown error";
    }
}

/* tcpip_read_* and tcpip_write_*
 
   Purpose: These procedures send and receive data on the socket.
*/

int tcpip_read_byte (int fd, UINT_8 *number)
{
  UINT_8 netChar;   /* char in network format */

  if (tcpip_readn (fd, (char *) &netChar, sizeof(UINT_8)) < (int) sizeof(UINT_8)) {
    return -1;
  }
 
  *number = netChar; 
  return 0;
}

int tcpip_read_short (int fd, UINT_16 *number)
{
  UINT_16 netShort;  /* short integer in network format */

  if (tcpip_readn (fd, (char *) &netShort, sizeof(UINT_16)) < (int) sizeof(UINT_16)) {
    return -1;
  }

  *number = ntohs (netShort); 
  return 0;
} 

int tcpip_read_long (int fd, UINT_32 *number)
{
  UINT_32 netLong;   /* long integer in network format */

  if (tcpip_readn (fd, (char *) &netLong, sizeof(UINT_32)) < (int) sizeof(UINT_32)) {
    return -1;
  }

  *number = ntohl (netLong);
  return 0;
}

char *tcpip_read_string (int fd)
{
  char *string;
  UINT_32 length;

  if (tcpip_read_long (fd, &length) < 0) {
    return NULL;
  }

  if (length == 0) {
    return NULL;
  }

  string = new char [length];

  if (tcpip_readn (fd, string, length) < (int) length) {
    delete string;
    return NULL;
  }

  string[length - 1] = '\0';
  return string;
}
 
int tcpip_write_byte (int fd, UINT_8 number)
{
  UINT_8 netChar = number;

  if (tcpip_writen (fd, (char *) &netChar, sizeof(UINT_8)) < 0)
    return -1;
 
  return 0;
}

int tcpip_write_short (int fd, UINT_16 number)
{
  UINT_16 netShort = htons (number);

  if (tcpip_writen (fd, (char *) &netShort, sizeof(UINT_16)) < 0)
    return -1;

  return 0;
}

int tcpip_write_long (int fd, UINT_32 number)
{
  UINT_32 netLong = htonl (number);

  if (tcpip_writen (fd, (char *) &netLong, sizeof(UINT_32)) < 0)
    return -1;

  return 0;
}

int tcpip_write_string (int fd, char *string)
{
  UINT_32 length = (string == NULL) ? 0 : strlen(string) + 1;

  if (tcpip_write_long (fd, length) < 0)
    return -1;

  if ((length > 0) && (tcpip_writen (fd, string, length) < 0))
    return -1;

  return 0;
}

int tcpip_stringtoip (char *string, UINT_32 &ip)
{
    if ((ip = (UINT_32) inet_addr(string)) != INADDR_NONE) {
	return 0;
    }

    return -1;
}

char *tcpip_iptostring (UINT_32 ip)
{
    struct in_addr inaddr;
    inaddr.s_addr = ip;
    return (my_strcpy (inet_ntoa (inaddr)));
}

int tcpip_reset_error (int sockfd)
{
   int rc;
   int optvalue;
   int optlength = sizeof(int);

   if ((rc = getsockopt (sockfd, SOL_SOCKET, SO_ERROR, (char *) &optvalue, &optlength)) < 0) {
	return rc;
   }

   return (optvalue);
}
